@@ -23,7 +23,6 @@
 use strict;
 use warnings;
 
-use Cwd;
 use File::Temp qw(tempdir);
 use File::Basename;
 use File::Copy;
@@ -38,12 +37,14 @@ use Dpkg::BuildProfiles qw(set_build_profiles);
 use Dpkg::Conf;
 use Dpkg::Compression;
 use Dpkg::Checksums;
+use Dpkg::Package;
 use Dpkg::Version;
 use Dpkg::Control;
 use Dpkg::Control::Info;
 use Dpkg::Changelog::Parse;
 use Dpkg::Path qw(find_command);
 use Dpkg::IPC;
+use Dpkg::Vendor qw(run_vendor_hook);
 
 textdomain('dpkg-dev');
 
@@ -63,21 +64,25 @@ sub usage {
 'Options:
       --build=<type>[,...]    specify the build <type>: full, source, binary,
                                 any, all (default is \'full\').
-  -F                          normal full build (source and binary; default).
-  -g                          source and arch-indep build.
-  -G                          source and arch-specific build.
-  -b                          binary-only, no source files.
-  -B                          binary-only, only arch-specific files.
-  -A                          binary-only, only arch-indep files.
-  -S                          source-only, no binary files.
+  -F, --build=full            normal full build (source and binary; default).
+  -g, --build=source,all      source and arch-indep build.
+  -G, --build=source,any      source and arch-specific build.
+  -b, --build=binary          binary-only, no source files.
+  -B, --build=any             binary-only, only arch-specific files.
+  -A, --build=all             binary-only, only arch-indep files.
+  -S, --build=source          source-only, no binary files.
   -nc, --no-pre-clean         do not pre clean source tree (implies -b).
-       --pre-clean            pre clean source tree (default).
-  -tc, --post-clean           clean source tree when finished.
-  -D                          check build dependencies and conflicts (default).
-  -d                          do not check build dependencies and conflicts.
-      --[no-]check-builddeps  ditto.
+      --pre-clean             pre clean source tree (default).
+      --no-post-clean         do not post clean source tree (default).
+  -tc, --post-clean           post clean source tree.
+      --sanitize-env          sanitize the build environment.
+  -D, --check-builddeps       check build dependencies and conflicts (default).
+  -d, --no-check-builddeps    do not check build dependencies and conflicts.
+      --ignore-builtin-builddeps
+                              do not check builtin build dependencies.
   -P, --build-profiles=<profiles>
-                              assume comma-separated build profiles as active.
+                              assume comma-separated build <profiles> as active.
+      --rules-requires-root   assume legacy Rules-Requires-Root field value.
   -R, --rules-file=<rules>    rules file to execute (default is debian/rules).
   -T, --rules-target=<target> call debian/rules <target>.
       --as-root               ensure -T calls the target with root rights.
@@ -85,22 +90,28 @@ sub usage {
                                 forced mode.
   -J, --jobs-try[=<number>|auto]
                               jobs to run simultaneously (passed to <rules>),
-                                opt-in mode.
+                                opt-in mode (default is auto).
   -r, --root-command=<command>
                               command to gain root rights (default is fakeroot).
       --check-command=<command>
                               command to check the .changes file (no default).
       --check-option=<opt>    pass <opt> to check <command>.
       --hook-<name>=<command> set <command> as the hook <name>, known hooks:
-                                init preclean source build binary changes
-                                postclean check sign done
+                                init preclean source build binary buildinfo
+                                changes postclean check sign done
+      --buildinfo-file=<file> set the .buildinfo filename to generate.
+      --buildinfo-option=<opt>
+                              pass option <opt> to dpkg-genbuildinfo.
+      --changes-file=<file>   set the .changes filename to generate.
   -p, --sign-command=<command>
                               command to sign .dsc and/or .changes files
-                                (default is gpg2 or gpg).
+                                (default is gpg).
   -k, --sign-key=<keyid>      the key to use for signing.
   -ap, --sign-pause           add pause before starting signature process.
   -us, --unsigned-source      unsigned source package.
-  -uc, --unsigned-changes     unsigned .changes file.
+  -ui, --unsigned-buildinfo   unsigned .buildinfo file.
+  -uc, --unsigned-changes     unsigned .buildinfo and .changes file.
+      --no-sign               do not sign any file.
       --force-sign            force signing the resulting files.
       --admindir=<directory>  change the administrative directory.
   -?, --help                  show this help message.
@@ -140,10 +151,11 @@ my $admindir;
 my @debian_rules = ('debian/rules');
 my @rootcommand = ();
 my $signcommand;
-my $noclean;
-my $cleansource;
+my $preclean = 1;
+my $postclean = 0;
+my $sanitize_env = 0;
 my $parallel;
-my $parallel_force;
+my $parallel_force = 0;
 my $checkbuilddep = 1;
 my $check_builtin_builddep = 1;
 my @source_opts;
@@ -154,6 +166,7 @@ my $signkey = $ENV{DEB_SIGN_KEYID};
 my $signforce = 0;
 my $signreleased = 1;
 my $signsource = 1;
+my $signbuildinfo = 1;
 my $signchanges = 1;
 my $buildtarget = 'build';
 my $binarytarget = 'binary';
@@ -162,15 +175,25 @@ my $host_type = '';
 my $target_arch = '';
 my $target_type = '';
 my @build_profiles = ();
-my $call_target = '';
+my $rrr_override;
+my @call_target = ();
 my $call_target_as_root = 0;
 my $since;
 my $maint;
 my $changedby;
 my $desc;
+my $buildinfo_file;
+my @buildinfo_opts;
+my $changes_file;
 my @changes_opts;
+my %target_legacy_root = map { $_ => 1 } qw(
+    clean binary binary-arch binary-indep
+);
+my %target_official =  map { $_ => 1 } qw(
+    clean build build-arch build-indep binary binary-arch binary-indep
+);
 my @hook_names = qw(
-    init preclean source build binary changes postclean check sign done
+    init preclean source build binary buildinfo changes postclean check sign done
 );
 my %hook;
 $hook{$_} = undef foreach @hook_names;
@@ -184,6 +207,8 @@ unshift @ARGV, @{$conf};
 
 my $build_opts = Dpkg::BuildOptions->new();
 
+run_vendor_hook('update-buildopts', $build_opts);
+
 if ($build_opts->has('nocheck')) {
     $check_command = undef;
 } elsif (not find_command($check_command)) {
@@ -205,8 +230,30 @@ while (@ARGV) {
 	$admindir = $1;
     } elsif (/^--source-option=(.*)$/) {
 	push @source_opts, $1;
+    } elsif (/^--buildinfo-file=(.*)$/) {
+        $buildinfo_file = $1;
+        usageerr(g_('missing .buildinfo filename')) if not length $buildinfo_file;
+    } elsif (/^--buildinfo-option=(.*)$/) {
+        my $buildinfo_opt = $1;
+        if ($buildinfo_opt =~ m/^-O(.*)/) {
+            warning(g_('passing %s via %s is not supported; please use %s instead'),
+                    '-O', '--buildinfo-option', '--buildinfo-file');
+            $buildinfo_file = $1;
+        } else {
+            push @buildinfo_opts, $buildinfo_opt;
+        }
+    } elsif (/^--changes-file=(.*)$/) {
+        $changes_file = $1;
+        usageerr(g_('missing .changes filename')) if not length $changes_file;
     } elsif (/^--changes-option=(.*)$/) {
-	push @changes_opts, $1;
+        my $changes_opt = $1;
+        if ($changes_opt =~ m/^-O(.*)/) {
+            warning(g_('passing %s via %s is not supported; please use %s instead'),
+                    '-O', '--changes-option', '--changes-file');
+            $changes_file = $1;
+        } else {
+            push @changes_opts, $changes_opt;
+        }
     } elsif (/^(?:-j|--jobs=)(\d*|auto)$/) {
 	$parallel = $1 || '';
 	$parallel_force = 1;
@@ -215,18 +262,21 @@ while (@ARGV) {
 	$parallel_force = 0;
     } elsif (/^(?:-r|--root-command=)(.*)$/) {
 	my $arg = $1;
-	@rootcommand = split /\s+/, $arg;
+	@rootcommand = split ' ', $arg;
     } elsif (/^--check-command=(.*)$/) {
 	$check_command = $1;
     } elsif (/^--check-option=(.*)$/) {
 	push @check_opts, $1;
-    } elsif (/^--hook-(.+)=(.*)$/) {
+    } elsif (/^--hook-([^=]+)=(.*)$/) {
 	my ($hook_name, $hook_cmd) = ($1, $2);
 	usageerr(g_('unknown hook name %s'), $hook_name)
 	    if not exists $hook{$hook_name};
 	usageerr(g_('missing hook %s command'), $hook_name)
 	    if not defined $hook_cmd;
 	$hook{$hook_name} = $hook_cmd;
+    } elsif (/^(--buildinfo-id)=.*$/) {
+	# Deprecated option
+	warning(g_('%s is deprecated; it is without effect'), $1);
     } elsif (/^(?:-p|--sign-command=)(.*)$/) {
 	$signcommand = $1;
     } elsif (/^(?:-k|--sign-key=)(.*)$/) {
@@ -242,9 +292,17 @@ while (@ARGV) {
 	warning(g_('-s%s is deprecated; always using gpg style interface'), $1);
     } elsif (/^--force-sign$/) {
 	$signforce = 1;
+    } elsif (/^--no-sign$/) {
+	$signforce = 0;
+	$signsource = 0;
+	$signbuildinfo = 0;
+	$signchanges = 0;
     } elsif (/^-us$/ or /^--unsigned-source$/) {
 	$signsource = 0;
+    } elsif (/^-ui$/ or /^--unsigned-buildinfo$/) {
+	$signbuildinfo = 0;
     } elsif (/^-uc$/ or /^--unsigned-changes$/) {
+	$signbuildinfo = 0;
 	$signchanges = 0;
     } elsif (/^-ap$/ or /^--sign-pausa$/) {
 	$signpause = 1;
@@ -264,7 +322,11 @@ while (@ARGV) {
     } elsif (/^-(?:s[nsAkurKUR]|[zZ].*|i.*|I.*)$/) {
 	push @source_opts, $_; # passed to dpkg-source
     } elsif (/^-tc$/ or /^--post-clean$/) {
-	$cleansource = 1;
+        $postclean = 1;
+    } elsif (/^--no-post-clean$/) {
+        $postclean = 0;
+    } elsif (/^--sanitize-env$/) {
+        $sanitize_env = 1;
     } elsif (/^-t$/ or /^--host-type$/) {
 	$host_type = shift; # Order DOES matter!
     } elsif (/^-t(.*)$/ or /^--host-type=(.*)$/) {
@@ -278,15 +340,18 @@ while (@ARGV) {
     } elsif (/^--target-type=(.*)$/) {
 	$target_type = $1;
     } elsif (/^(?:--target|--rules-target|-T)$/) {
-        $call_target = shift @ARGV;
+        push @call_target, split /,/, shift @ARGV;
     } elsif (/^(?:--target=|--rules-target=|-T)(.+)$/) {
-        $call_target = $1;
+        my $arg = $1;
+        push @call_target, split /,/, $arg;
+    } elsif (/^--rules-requires-root$/) {
+        $rrr_override = 'binary-targets';
     } elsif (/^--as-root$/) {
         $call_target_as_root = 1;
     } elsif (/^--pre-clean$/) {
-	$noclean = 0;
+        $preclean = 1;
     } elsif (/^-nc$/ or /^--no-pre-clean$/) {
-	$noclean = 1;
+        $preclean = 0;
     } elsif (/^--build=(.*)$/) {
         set_build_type_from_options($1, $_);
     } elsif (/^-b$/) {
@@ -313,15 +378,20 @@ while (@ARGV) {
 	$desc = $1;
     } elsif (m/^-[EW]$/) {
 	# Deprecated option
-	warning(g_('-E and -W are deprecated, they are without effect'));
-    } elsif (/^-R(.*)$/ or /^--rules-target=(.*)$/) {
+	warning(g_('%s is deprecated; it is without effect'), $_);
+    } elsif (/^-R(.*)$/ or /^--rules-file=(.*)$/) {
 	my $arg = $1;
-	@debian_rules = split /\s+/, $arg;
+	@debian_rules = split ' ', $arg;
     } else {
 	usageerr(g_('unknown option or argument %s'), $_);
     }
 }
 
+if (@call_target) {
+    my $targets = join ',', @call_target;
+    set_build_type_from_targets($targets, '--rules-target', nocheck => 1);
+}
+
 if (build_has_all(BUILD_BINARY)) {
     $buildtarget = 'build';
     $binarytarget = 'binary';
@@ -333,52 +403,47 @@ if (build_has_all(BUILD_BINARY)) {
     $binarytarget = 'binary-indep';
 }
 
-if ($noclean) {
+if (not $preclean) {
     # -nc without -b/-B/-A/-S/-F implies -b
     set_build_type(BUILD_BINARY) if build_has_any(BUILD_DEFAULT);
     # -nc with -S implies no dependency checks
     $checkbuilddep = 0 if build_is(BUILD_SOURCE);
 }
 
-if ($< == 0) {
-    warning(g_('using a gain-root-command while being root')) if (@rootcommand);
-} else {
-    push @rootcommand, 'fakeroot' unless @rootcommand;
-}
-
-if (@rootcommand and not find_command($rootcommand[0])) {
-    if ($rootcommand[0] eq 'fakeroot' and $< != 0) {
-        error(g_("fakeroot not found, either install the fakeroot\n" .
-                 'package, specify a command with the -r option, ' .
-                 'or run this as root'));
-    } else {
-        error(g_("gain-root-commmand '%s' not found"), $rootcommand[0]);
-    }
+if ($call_target_as_root and @call_target == 0) {
+    error(g_('option %s is only meaningful with option %s'),
+          '--as-root', '--rules-target');
 }
 
 if ($check_command and not find_command($check_command)) {
-    error(g_("check-commmand '%s' not found"), $check_command);
+    error(g_("check-command '%s' not found"), $check_command);
 }
 
 if ($signcommand) {
     if (!find_command($signcommand)) {
-        error(g_("sign-commmand '%s' not found"), $signcommand);
+        error(g_("sign-command '%s' not found"), $signcommand);
     }
 } elsif (($ENV{GNUPGHOME} && -e $ENV{GNUPGHOME}) ||
          ($ENV{HOME} && -e "$ENV{HOME}/.gnupg")) {
-    if (find_command('gpg2')) {
-        $signcommand = 'gpg2';
-    } elsif (find_command('gpg')) {
+    if (find_command('gpg')) {
         $signcommand = 'gpg';
     }
 }
 
+# Default to auto if none of parallel=N, -J or -j have been specified.
+if (not defined $parallel and not $build_opts->has('parallel')) {
+    $parallel = 'auto';
+}
+
 if (defined $parallel) {
     if ($parallel eq 'auto') {
         # Most Unices.
         $parallel = qx(getconf _NPROCESSORS_ONLN 2>/dev/null);
         # Fallback for at least Irix.
         $parallel = qx(getconf _NPROC_ONLN 2>/dev/null) if $?;
+        # Fallback to serial execution if cannot infer the number of online
+        # processors.
+        $parallel = '1' if $?;
         chomp $parallel;
     }
     if ($parallel_force) {
@@ -387,14 +452,23 @@ if (defined $parallel) {
     }
     $build_opts->set('parallel', $parallel);
     $build_opts->export();
+} else {
+    $build_opts->export();
 }
 
-set_build_profiles(@build_profiles) if @build_profiles;
+if ($build_opts->has('terse')) {
+    $ENV{MAKEFLAGS} //= '';
+    $ENV{MAKEFLAGS} .= ' --no-print-directory';
+}
 
-my $cwd = cwd();
-my $dir = basename($cwd);
+set_build_profiles(@build_profiles);
 
 my $changelog = changelog_parse();
+my $ctrl = Dpkg::Control::Info->new();
+
+# Check whether we are doing some kind of rootless build, and sanity check
+# the fields values.
+my %rules_requires_root = parse_rules_requires_root($ctrl->get_source());
 
 my $pkg = mustsetvar($changelog->{source}, g_('source package'));
 my $version = mustsetvar($changelog->{version}, g_('source version'));
@@ -417,7 +491,7 @@ if ($changedby) {
 }
 
 # <https://reproducible-builds.org/specs/source-date-epoch/>
-$ENV{SOURCE_DATE_EPOCH} ||= $changelog->{timestamp};
+$ENV{SOURCE_DATE_EPOCH} ||= $changelog->{timestamp} || time;
 
 my @arch_opts;
 push @arch_opts, ('--host-arch', $host_arch) if $host_arch;
@@ -446,15 +520,21 @@ if (build_has_any(BUILD_ARCH_DEP)) {
 my $pv = "${pkg}_$sversion";
 my $pva = "${pkg}_${sversion}_$arch";
 
+signkey_validate();
+
 if (not $signcommand) {
     $signsource = 0;
+    $signbuildinfo = 0;
     $signchanges = 0;
 } elsif ($signforce) {
     $signsource = 1;
+    $signbuildinfo = 1;
     $signchanges = 1;
-} elsif (($signsource or $signchanges) and $distribution eq 'UNRELEASED') {
+} elsif (($signsource or $signbuildinfo or $signchanges) and
+         $distribution eq 'UNRELEASED') {
     $signreleased = 0;
     $signsource = 0;
+    $signbuildinfo = 0;
     $signchanges = 0;
 }
 
@@ -462,6 +542,11 @@ if ($signsource && build_has_none(BUILD_SOURCE)) {
     $signsource = 0;
 }
 
+# Sanitize build environment.
+if ($sanitize_env) {
+    run_vendor_hook('sanitize-environment');
+}
+
 #
 # Preparation of environment stops here
 #
@@ -473,10 +558,8 @@ if (not -x 'debian/rules') {
     chmod(0755, 'debian/rules'); # No checks of failures, non fatal
 }
 
-unless ($call_target) {
-    chdir('..') or syserr('chdir ..');
-    withecho('dpkg-source', @source_opts, '--before-build', $dir);
-    chdir($dir) or syserr("chdir $dir");
+if (scalar @call_target == 0) {
+    run_cmd('dpkg-source', @source_opts, '--before-build', '.');
 }
 
 if ($checkbuilddep) {
@@ -497,112 +580,124 @@ if ($checkbuilddep) {
     }
 }
 
-if ($call_target) {
-    if ($call_target_as_root or
-        $call_target =~ /^(clean|binary(|-arch|-indep))$/)
-    {
-        withecho(@rootcommand, @debian_rules, $call_target);
-    } else {
-        withecho(@debian_rules, $call_target);
-    }
-    exit 0;
+foreach my $call_target (@call_target) {
+    run_rules_cond_root($call_target);
 }
+exit 0 if scalar @call_target;
 
-run_hook('preclean', ! $noclean);
+run_hook('preclean', $preclean);
 
-unless ($noclean) {
-    withecho(@rootcommand, @debian_rules, 'clean');
+if ($preclean) {
+    run_rules_cond_root('clean');
 }
 
 run_hook('source', build_has_any(BUILD_SOURCE));
 
 if (build_has_any(BUILD_SOURCE)) {
     warning(g_('building a source package without cleaning up as you asked; ' .
-               'it might contain undesired files')) if $noclean;
-    chdir('..') or syserr('chdir ..');
-    withecho('dpkg-source', @source_opts, '-b', $dir);
-    chdir($dir) or syserr("chdir $dir");
+               'it might contain undesired files')) if not $preclean;
+    run_cmd('dpkg-source', @source_opts, '-b', '.');
 }
 
 run_hook('build', build_has_any(BUILD_BINARY));
 
-# XXX Use some heuristics to decide whether to use build-{arch,indep} targets.
-# This is a temporary measure to not break too many packages on a flag day.
-build_target_fallback();
+my $build_types = get_build_options_from_type();
 
 if (build_has_any(BUILD_BINARY)) {
-    withecho(@debian_rules, $buildtarget);
+    # XXX Use some heuristics to decide whether to use build-{arch,indep}
+    # targets. This is a temporary measure to not break too many packages
+    # on a flag day.
+    build_target_fallback($ctrl);
+
+    # If we are building rootless, there is no need to call the build target
+    # independently as non-root.
+    run_cmd(@debian_rules, $buildtarget) if rules_requires_root($binarytarget);
     run_hook('binary', 1);
-    withecho(@rootcommand, @debian_rules, $binarytarget);
+    run_rules_cond_root($binarytarget);
 }
 
+run_hook('buildinfo', 1);
+
+$buildinfo_file //= "../$pva.buildinfo";
+
+push @buildinfo_opts, "--build=$build_types" if build_has_none(BUILD_DEFAULT);
+push @buildinfo_opts, "--admindir=$admindir" if $admindir;
+push @buildinfo_opts, "-O$buildinfo_file" if $buildinfo_file;
+
+run_cmd('dpkg-genbuildinfo', @buildinfo_opts);
+
 run_hook('changes', 1);
 
-my $build_types = get_build_options_from_type();
+$changes_file //= "../$pva.changes";
+
 push @changes_opts, "--build=$build_types" if build_has_none(BUILD_DEFAULT);
 push @changes_opts, "-m$maint" if defined $maint;
 push @changes_opts, "-e$changedby" if defined $changedby;
 push @changes_opts, "-v$since" if defined $since;
 push @changes_opts, "-C$desc" if defined $desc;
+push @changes_opts, "-O$changes_file";
 
-my $chg = "../$pva.changes";
 my $changes = Dpkg::Control->new(type => CTRL_FILE_CHANGES);
 
-printcmd("dpkg-genchanges @changes_opts >$chg");
+run_cmd('dpkg-genchanges', @changes_opts);
+$changes->load($changes_file);
 
-open my $changes_fh, '-|', 'dpkg-genchanges', @changes_opts
-    or subprocerr('dpkg-genchanges');
-$changes->parse($changes_fh, g_('parse changes file'));
-$changes->save($chg);
-close $changes_fh or subprocerr(g_('dpkg-genchanges'));
+run_hook('postclean', $postclean);
 
-run_hook('postclean', $cleansource);
-
-if ($cleansource) {
-    withecho(@rootcommand, @debian_rules, 'clean');
+if ($postclean) {
+    run_rules_cond_root('clean');
 }
 
-chdir('..') or syserr('chdir ..');
-withecho('dpkg-source', @source_opts, '--after-build', $dir);
-chdir($dir) or syserr("chdir $dir");
+run_cmd('dpkg-source', @source_opts, '--after-build', '.');
 
 info(describe_build($changes->{'Files'}));
 
 run_hook('check', $check_command);
 
 if ($check_command) {
-    withecho($check_command, @check_opts, $chg);
+    run_cmd($check_command, @check_opts, $changes_file);
 }
 
-if ($signpause && ($signchanges || $signsource)) {
+if ($signpause && ($signsource || $signbuildinfo || $signchanges)) {
     print g_("Press <enter> to start the signing process.\n");
     getc();
 }
 
-run_hook('sign', $signsource || $signchanges);
+run_hook('sign', $signsource || $signbuildinfo || $signchanges);
 
 if ($signsource) {
     if (signfile("$pv.dsc")) {
-        error(g_('failed to sign .dsc and .changes file'));
+        error(g_('failed to sign %s file'), '.dsc');
     }
 
     # Recompute the checksums as the .dsc has changed now.
+    my $buildinfo = Dpkg::Control->new(type => CTRL_FILE_BUILDINFO);
+    $buildinfo->load($buildinfo_file);
+    my $checksums = Dpkg::Checksums->new();
+    $checksums->add_from_control($buildinfo);
+    $checksums->add_from_file("../$pv.dsc", update => 1, key => "$pv.dsc");
+    $checksums->export_to_control($buildinfo);
+    $buildinfo->save($buildinfo_file);
+}
+if ($signbuildinfo && signfile("$pva.buildinfo")) {
+    error(g_('failed to sign %s file'), '.buildinfo');
+}
+if ($signsource or $signbuildinfo) {
+    # Recompute the checksums as the .dsc and/or .buildinfo have changed.
     my $checksums = Dpkg::Checksums->new();
     $checksums->add_from_control($changes);
-    $checksums->add_from_file("../$pv.dsc", update => 1, key => "$pv.dsc");
+    $checksums->add_from_file("../$pv.dsc", update => 1, key => "$pv.dsc")
+        if $signsource;
+    $checksums->add_from_file($buildinfo_file, update => 1, key => "$pva.buildinfo");
     $checksums->export_to_control($changes);
     delete $changes->{'Checksums-Md5'};
-
-    my $md5sum_regex = checksums_get_property('md5', 'regex');
-    my $dsc_md5sum = $checksums->get_checksum("$pv.dsc", 'md5');
-    my $dsc_size = $checksums->get_size("$pv.dsc");
-    my $dsc_files_regex = qr/$md5sum_regex\s+\d+\s+(\S+\s+\S+\s+\Q$pv\E\.dsc)/;
-    $changes->{'Files'} =~ s/^$dsc_files_regex$/$dsc_md5sum $dsc_size $1/m;
-
-    $changes->save($chg);
+    update_files_field($changes, $checksums, "$pv.dsc")
+        if $signsource;
+    update_files_field($changes, $checksums, "$pva.buildinfo");
+    $changes->save($changes_file);
 }
 if ($signchanges && signfile("$pva.changes")) {
-    error(g_('failed to sign .changes file'));
+    error(g_('failed to sign %s file'), '.changes');
 }
 
 if (not $signreleased) {
@@ -621,10 +716,108 @@ sub mustsetvar {
     return $var;
 }
 
-sub withecho {
+sub setup_rootcommand {
+    if ($< == 0) {
+        warning(g_('using a gain-root-command while being root')) if @rootcommand;
+    } else {
+        push @rootcommand, 'fakeroot' unless @rootcommand;
+    }
+
+    if (@rootcommand and not find_command($rootcommand[0])) {
+        if ($rootcommand[0] eq 'fakeroot' and $< != 0) {
+            error(g_("fakeroot not found, either install the fakeroot\n" .
+                     'package, specify a command with the -r option, ' .
+                     'or run this as root'));
+        } else {
+            error(g_("gain-root-command '%s' not found"), $rootcommand[0]);
+        }
+    }
+}
+
+sub parse_rules_requires_root {
+    my $ctrl = shift;
+
+    my %rrr;
+    my $rrr;
+    my $keywords_base;
+    my $keywords_impl;
+
+    $rrr = $rrr_override // $ctrl->{'Rules-Requires-Root'} // 'binary-targets';
+
+    foreach my $keyword (split ' ', $rrr) {
+        if ($keyword =~ m{/}) {
+            if ($keyword =~ m{^dpkg/target/(.*)$}p and $target_official{$1}) {
+                error(g_('disallowed target in %s field keyword "%s"'),
+                      'Rules-Requires-Root', $keyword);
+            } elsif ($keyword ne 'dpkg/target-subcommand') {
+                error(g_('%s field keyword "%s" is unknown in dpkg namespace'),
+                      'Rules-Requires-Root', $keyword);
+            }
+            $keywords_impl++;
+        } else {
+            if ($keyword ne lc $keyword and
+                (lc $keyword eq 'no' or lc $keyword eq 'binary-targets')) {
+                error(g_('%s field keyword "%s" is uppercase; use "%s" instead'),
+                      'Rules-Requires-Root', $keyword, lc $keyword);
+            } elsif (lc $keyword eq 'yes') {
+                error(g_('%s field keyword "%s" is invalid; use "%s" instead'),
+                      'Rules-Requires-Root', $keyword, 'binary-targets');
+            } elsif ($keyword ne 'no' and $keyword ne 'binary-targets') {
+                warning(g_('%s field keyword "%s" is unknown'),
+                        'Rules-Requires-Root', $keyword);
+            }
+            $keywords_base++;
+        }
+
+        if ($rrr{$keyword}++) {
+            error(g_('field %s contains duplicate keyword "%s"'),
+                        'Rules-Requires-Root', $keyword);
+        }
+    }
+
+    if ($call_target_as_root or not exists $rrr{no}) {
+        setup_rootcommand();
+    }
+
+    # Notify the children we do support R³.
+    $ENV{DEB_RULES_REQUIRES_ROOT} = join ' ', sort keys %rrr;
+
+    if ($keywords_base > 1 or $keywords_base and $keywords_impl) {
+        error(g_('%s field contains both global and implementation specific keywords'),
+              'Rules-Requires-Root');
+    } elsif ($keywords_impl) {
+        # Set only on <implementations-keywords>.
+        $ENV{DEB_GAIN_ROOT_CMD} = join ' ', @rootcommand;
+    } else {
+        # We should not provide the variable otherwise.
+        delete $ENV{DEB_GAIN_ROOT_CMD};
+    }
+
+    return %rrr;
+}
+
+sub run_cmd {
     printcmd(@_);
-    system(@_)
-	and subprocerr("@_");
+    system @_ and subprocerr("@_");
+}
+
+sub rules_requires_root {
+    my $target = shift;
+
+    return 1 if $call_target_as_root;
+    return 1 if $rules_requires_root{"dpkg/target/$target"};
+    return 1 if $rules_requires_root{'binary-targets'} and $target_legacy_root{$target};
+    return 0;
+}
+
+sub run_rules_cond_root {
+    my $target = shift;
+
+    my @cmd;
+    push @cmd, @rootcommand if rules_requires_root($target);
+    push @cmd, @debian_rules, $target;
+
+    run_cmd(@cmd);
 }
 
 sub run_hook {
@@ -655,9 +848,38 @@ sub run_hook {
         }
     };
 
-    $cmd =~ s/\%(.)/&$subst_hook_var($1)/eg;
+    $cmd =~ s/\%(.)/$subst_hook_var->($1)/eg;
 
-    withecho($cmd);
+    run_cmd($cmd);
+}
+
+sub update_files_field {
+    my ($ctrl, $checksums, $filename) = @_;
+
+    my $md5sum_regex = checksums_get_property('md5', 'regex');
+    my $md5sum = $checksums->get_checksum($filename, 'md5');
+    my $size = $checksums->get_size($filename);
+    my $file_regex = qr/$md5sum_regex\s+\d+\s+(\S+\s+\S+\s+\Q$filename\E)/;
+
+    $ctrl->{'Files'} =~ s/^$file_regex$/$md5sum $size $1/m;
+}
+
+sub signkey_validate {
+    return unless defined $signkey;
+    # Make sure this is an hex keyid.
+    return unless $signkey =~ m/^(?:0x)?([[:xdigit:]]+)$/;
+
+    my $keyid = $1;
+
+    if (length $keyid <= 8) {
+        error(g_('short OpenPGP key IDs are broken; ' .
+                 'please use key fingerprints in %s or %s instead'),
+              '-k', 'DEB_SIGN_KEYID');
+    } elsif (length $keyid <= 16) {
+        warning(g_('long OpenPGP key IDs are strongly discouraged; ' .
+                   'please use key fingerprints in %s or %s instead'),
+                '-k', 'DEB_SIGN_KEYID');
+    }
 }
 
 sub signfile {
@@ -676,11 +898,12 @@ sub signfile {
 
     system($signcommand, '--utf8-strings', '--textmode', '--armor',
            '--local-user', $signkey || $maintainer, '--clearsign',
+           '--weak-digest', 'SHA1', '--weak-digest', 'RIPEMD160',
            '--output', "$signfile.asc", $signfile);
     my $status = $?;
     if ($status == 0) {
-	system('mv', '--', "$signfile.asc", "../$file")
-	    and subprocerr('mv');
+        move("$signfile.asc", "../$file")
+            or syserror(g_('cannot move %s to %s'), "$signfile.asc", "../$file");
     }
 
     print "\n";
@@ -690,7 +913,7 @@ sub signfile {
 sub fileomitted {
     my ($files, $regex) = @_;
 
-    return $files !~ /$regex/
+    return $files !~ m/$regex$/m
 }
 
 sub describe_build {
@@ -720,13 +943,18 @@ sub describe_build {
 }
 
 sub build_target_fallback {
+    my $ctrl = shift;
+
+    # If we are building rootless, there is no need to call the build target
+    # independently as non-root.
+    return if not rules_requires_root($binarytarget);
+
     return if $buildtarget eq 'build';
     return if scalar @debian_rules != 1;
 
     # Check if we are building both arch:all and arch:any packages, in which
     # case we now require working build-indep and build-arch targets.
     my $pkg_arch = 0;
-    my $ctrl = Dpkg::Control::Info->new();
 
     foreach my $bin ($ctrl->get_packages()) {
         if ($bin->{Architecture} eq 'all') {
@@ -740,7 +968,7 @@ sub build_target_fallback {
 
     # Check if the build-{arch,indep} targets are supported. If not, fallback
     # to build.
-    my $pid = spawn(exec => [ 'make', '-f', @debian_rules, '-qn', $buildtarget ],
+    my $pid = spawn(exec => [ $Dpkg::PROGMAKE, '-f', @debian_rules, '-qn', $buildtarget ],
                     from_file => '/dev/null', to_file => '/dev/null',
                     error_to_file => '/dev/null');
     my $cmdline = "make -f @debian_rules -qn $buildtarget";
